#include "config.h"
#include "err.h"
#include "builtin.h"
#include "welltype/version.h"
#include "welltype/wtruntime.h"
#include "welltype/libwtref.h"
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>

static void print_help(const char* progname)
{
	printf("Usage: %s [OPTION]... FILE\n", progname);
	printf("\n");
	printf("Welltype runtime environment.\n");
	printf("\n");
	printf("    -b        Add runtime's builtins\n");
	printf("    -i        Load only integrity verified binaries\n");
	printf("    -l FILE   Load library from FILE\n");
	printf("    -M KEY    Print metadata value by KEY and exit\n");
/*	Printf("    -s SIZE   Specify memory size\n"); */
	printf("    -v        Print version information and exit\n");
}

static void* alloc_func(void* priv, void* mem, size_t size)
{
	(void)priv;
	return realloc(mem, size);
}

static wt_program load_program(const char* const progname, struct wtr_vm* const vm,
	struct wtr_context* const ctx, const wt_u8* data, size_t size,
	const wt_u32 flags)
{
	wt_program prog;
	wt_error   err;
	wt_u32     retc;

	if(2<size && 0==memcmp(data, "#!", 2))
	{
		size -= 2;
		data += 2;

		/* skip the shebang (find the magic) */

		while(sizeof(wt_u32)<size)
		{
			if(WTBMAG0==data[0] && WTBMAG1==data[1] && WTBMAG2==data[2] &&
				WTBMAG3==data[3])
			{
				break;
			}

			++data;
			--size;
		}
	}

	if(WTE_OK!=(err=wtrVMLoadProgram(vm, &prog, ctx, data, size, flags)))
	{
		wt_display_error(progname, "failed to load program", err);
		return NULL;
	}

	if(!libwtref_run_vm(vm, ctx, wt_error_cb, (void*)progname, &retc, 0,
		NULL))
	{
		wtrVMReleaseProgram(vm, prog, ctx);
		libwtref_run_vm(vm, ctx, wt_error_cb, (void*)progname, &retc, 0,
			NULL);
		return NULL;
	}

	return prog;
}

static void unload_program(const char* const progname, struct wtr_vm* const vm,
	wt_program program, struct wtr_context* const ctx)
{
	wt_u32 retc;
	wtrVMReleaseProgram(vm, program, ctx);
	libwtref_run_vm(vm, ctx, wt_error_cb, (void*)progname, &retc, 0, NULL);
}

static wt_program read_program(const char* const progname, struct wtr_vm* const vm,
	struct wtr_context* const ctx, const char* const filename,
	const wt_u32 flags)
{
	wt_program prog;
	FILE*      file;
	wt_u8*     data;
	size_t     size;

	if(NULL==(file=fopen(filename, "rb")))
	{
		const int err = errno;
		fprintf(stderr, "%s: failed to open file '%s': %s\n", progname, filename, strerror(err));
		return NULL;
	}

	/* TODO: maybe replace with fstat(fileno(file)) */
	/* but we need to detect fileno(3) for this */
	/* for now, we use the original mechanism */
	fseek(file, 0, SEEK_END);
	size = (size_t)ftell(file);
	fseek(file, 0, SEEK_SET);

	if(NULL==(data=(wt_u8*)malloc(size)))
	{
		fclose(file);
		/* TODO: use %zu is possible! */
		fprintf(stderr, "%s: could not allocate %lu bytes of memory\n", progname, (unsigned long)size);
		return NULL;
	}

	if(1!=fread(data, size, 1, file))
	{
		const int err = errno;
		fclose(file);
		free(data);
		fprintf(stderr, "%s: failed to read file '%s': %s\n", progname, filename, strerror(err));
		return NULL;
	}

	fclose(file);

	prog = load_program(progname, vm, ctx, data, size, flags);
	free(data);
	return prog;
}

struct lib
{
	const char* name;
	wt_program  prog;
};

static wt_bool load_library(const char* const progname, struct wtr_vm* const vm,
	struct wtr_context* const ctx, struct lib* const lib, const wt_u32 flags)
{
	return NULL!=(lib->prog=read_program(progname, vm, ctx, lib->name, flags));
}

static wt_bool execute(const char* const progname, struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_program program)
{
	wt_error err;
	wt_u32   retc;

	if(WTE_OK!=(err=wtrContextInvokeProgram(ctx, program)))
	{
		wt_display_error(progname, "failed to invoke program", err);
		return WT_FALSE;
	}

	return libwtref_run_vm(vm, ctx, wt_error_cb, (void*)progname, &retc, 0,
		NULL);
}

int main(int argc, char* argv[])
{
	struct wtr_vm*    vm;
	wt_program        program;
	wt_error          err;
	wt_bool           addbltin = WT_FALSE;
	wt_u32            vm_flags = 0;
	wt_u32            l_flags  = 0;
	struct lib* const libvec   = (struct lib*)malloc(sizeof(struct lib)*argc);
	size_t            libcount = 0;
	size_t            megs     = 64;
	const char* const progname = argv[0];
	const char*       filename = NULL;
	const char*       metakey  = NULL;
	wt_bool           run_ok   = WT_TRUE;

	struct libwtref_segments segs;

	struct wtr_context* context;

	if(NULL==libvec)
	{
		fprintf(stderr, "%s: out of memory\n", progname);
		return 1;
	}

	++argv, --argc;

	while(argc)
	{
		if('-'==argv[0][0])
		{
			if(0==strcmp("--help", argv[0]))
			{
				print_help(progname);
				return 0;
			}
			else
			if(0==strcmp("-b", argv[0]))
			{
				addbltin = WT_TRUE;
			}
			else
			if(0==strcmp("-i", argv[0]))
			{
				vm_flags |= WTRVMF_REQINTEGRITY;
			}
			else
			if(0==strcmp("-v", argv[0]) || 0==strcmp("--version", argv[0]))
			{
				const wt_u32 wtver = wtrGetVersion();

				printf("Program:  v%u.%u.%u\n",
					(unsigned int)WT_GETMAJOR(WTR_VERSION),
					(unsigned int)WT_GETMINOR(WTR_VERSION),
					(unsigned int)WT_GETPATCH(WTR_VERSION)
					);

				printf("Welltype: v%u.%u.%u\n",
					(unsigned int)WT_GETMAJOR(wtver),
					(unsigned int)WT_GETMINOR(wtver),
					(unsigned int)WT_GETPATCH(wtver)
					);

				return 0;
			}
			else
			if(0==strcmp("-l", argv[0]))
			{
				if(1<argc)
				{
					libvec[libcount++].name = argv[1];
					++argv, --argc;
				}
				else
				{
					goto usage_and_exit;
				}
			}
			else
			if(0==strcmp("-M", argv[0]))
			{
				if(NULL==metakey && 1<argc)
				{
					metakey = argv[1];
					++argv, --argc;
				}
				else
				{
					goto usage_and_exit;
				}
			}
			else
			{
				fprintf(stderr, "%s: unknown switch: '%s'\n", progname, argv[0]);
				goto usage_and_exit;
			}
		}
		else
		if(filename)
		{
			fprintf(stderr, "%s: extra operand '%s'\n", progname, argv[0]);
			return 1;
		}
		else
		{
			filename = argv[0];
		}

		++argv, --argc;
	}

	if(!filename)
	{
usage_and_exit:
		fprintf(stderr, "Usage: %s [OPTION]... FILE\n", progname);
		return 1;
	}

	if(0!=libwtref_segments_create(&segs, megs*1024*1024, 0))
	{
		fprintf(stderr, "Failed to create memory segmentation.\n");
		return 1;
	}

	if(WTE_OK!=(err=wtrCreateTombVM(&vm, vm_flags, segs.vma_segs,
		segs.vma_nsegs, segs.vma_szsh, segs.heap_start, segs.heap_size,
		segs.heap_gran, 8, 2, NULL, alloc_func)))
	{
		wt_display_error(progname, "Failed to create the virtual machine",
			err);
		return 1;
	}

	/* the legacy default stack size is 10 KiB */
	if(WTE_OK!=(err=wtrContextCreate(&context, vm, 0, 10*1024, 0, 0)))
	{
		wt_display_error(progname, "Failed to create context", err);
		wtrVMDestroy(vm);
		return 1;
	}

	if(WTE_OK!=libwtref_init(vm, wt_error_cb, (void*)progname))
	{
		return 2;
	}

	if(WTE_OK!=libwtref_set_loader_ops(vm, stdout))
	{
		return 2;
	}

	if(addbltin && 0!=wt_add_builtins(progname, vm))
	{
		return 2;
	}

	{
		size_t i;

		for(i=0;i<libcount;++i)
		{
			if(!load_library(progname, vm, context, &libvec[i], l_flags))
			{
				return 3;
			}
		}
	}

	if(NULL==(program=read_program(progname, vm, context, filename, l_flags)))
	{
		return 3;
	}

	run_ok = execute(progname, vm, context, program);

	{
		wt_u32 retc;
		wtrContextCleanup(context);
		libwtref_run_vm(vm, context, wt_error_cb, (void*)progname, &retc, 0,
			NULL);
	}

	unload_program(progname, vm, program, context);
	while(0<libcount)
	{
		unload_program(progname, vm, libvec[--libcount].prog, context);
	}

	wtrContextDestroy(context);
	wtrVMDestroy(vm);

	libwtref_segments_destroy(&segs);
	free(libvec);

	if(!run_ok)
	{
		return 4;
	}

	return 0;
}
