//================================================================================================
// $Id: codegen.cpp 6535 2023-04-04 20:02:08Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "codegen.h"
#include "cgvisitor.h"
#include "cgsection.h"
#include "cgfile.h"
#include "parserext.h"
#include "framestack.h"
#include "validator.h"
#include "impincwriter.h"
#include <iostream>
#include <sstream>
#include <iomanip>
#include <cstring>
#include <cstdlib>
#include "welltype/version.h"
#include "welltype/wtruntime.h"
#include "wtruntime/types.h"
#include "wtruntime/hash.h"
#include "libwtref/libwtref.h"
#include "wtparser/astvisitor.h"
#include "wtparser/ast/ast_program.h"
#include "wtparser/ast/ast_entity.h"
#include "wtparser/ast/ast_metadata.h"
#include "wtparser/ast/ast_block.h"
#include "wtparser/ast/ast_mainblock.h"
#include "wtparser/ast/ast_constructorblock.h"
#include "wtparser/ast/ast_destructorblock.h"
#include "wtparser/ast/ast_declareblock.h"
#include "wtparser/ast/ast_compound.h"
#include "wtparser/ast/ast_exceptioncompound.h"
#include "wtparser/ast/ast_ehandler.h"
#include "wtparser/ast/ast_ehandlers.h"
#include "wtparser/ast/ast_finally.h"
#include "wtparser/ast/ast_statement.h"
#include "wtparser/ast/ast_variableaccess.h"
#include "wtparser/ast/ast_variabledeclaration.h"
#include "wtparser/ast/ast_variabledeclfragment.h"
#include "wtparser/ast/ast_elementaccess.h"
#include "wtparser/ast/ast_type.h"
#include "wtparser/ast/ast_typeinfo.h"
#include "wtparser/ast/ast_simultaneoustype.h"
#include "wtparser/ast/ast_basictype.h"
#include "wtparser/ast/ast_seqtype.h"
#include "wtparser/ast/ast_simultaneousexpression.h"
#include "wtparser/ast/ast_stringliteral.h"
#include "wtparser/ast/ast_booleanliteral.h"
#include "wtparser/ast/ast_characterliteral.h"
#include "wtparser/ast/ast_numberliteral.h"
#include "wtparser/ast/ast_assignment.h"
#include "wtparser/ast/ast_functionsignature.h"
#include "wtparser/ast/ast_functionargument.h"
#include "wtparser/ast/ast_functiondefinition.h"
#include "wtparser/ast/ast_functiondeclaration.h"
#include "wtparser/ast/ast_functioncall.h"
#include "wtparser/ast/ast_procedurecall.h"
#include "wtparser/ast/ast_recordfield.h"
#include "wtparser/ast/ast_recordtype.h"
#include "wtparser/ast/ast_recorddeclaration.h"
#include "wtparser/ast/ast_exceptiondeclaration.h"
#include "wtparser/ast/ast_precondition.h"
#include "wtparser/ast/ast_postcondition.h"
#include "wtparser/ast/ast_memberaccess.h"
#include "wtparser/ast/ast_andthen.h"
#include "wtparser/ast/ast_orelse.h"
#include "wtparser/ast/ast_returnstatement.h"
#include "wtparser/ast/ast_operatordeclaration.h"
#include "wtparser/ast/ast_raisestatement.h"
#include "wtparser/ast/ast_whilestatement.h"
#include "wtparser/ast/ast_dowhilestatement.h"
#include "wtparser/ast/ast_forstatement.h"
#include "wtparser/ast/ast_foreachstatement.h"
#include "wtparser/ast/ast_ifstatement.h"
#include "wtparser/ast/ast_globalblock.h"
#include "wtparser/ast/ast_assertstatement.h"
#include "wtparser/ast/ast_datatype.h"
#include "wtparser/ast/ast_datatypecons.h"
#include "wtparser/ast/ast_datatypedeclaration.h"
#include "wtparser/ast/ast_datatypeliteral.h"
#include "wtparser/ast/ast_functiontype.h"
#include "wtparser/ast/ast_functiontypedeclaration.h"
#include "wtparser/ast/ast_makefunctionpointer.h"
#include "wtparser/ast/ast_enumerator.h"
#include "wtparser/ast/ast_enumtype.h"
#include "wtparser/ast/ast_enumtypedeclaration.h"

namespace wtc
{
	bool AddAssert_impl(ProblemHandler* const probhan, const char* const file,
		unsigned int line, const char* const expr)
	{
		std::ostringstream oss;
		oss << "Assertion failed at " << file << ":" << line << " : " << expr;
		probhan->AddError("<internal>", 0, oss.str());
		return true;
	}

	static const struct wtr_hashalg* find_hash_alg(const char* const name)
	{
		for(const struct wtr_hashalg* alg=libwtref_hash_algs;
			NULL!=alg->name;++alg)
		{
			if(0==strcmp(name, alg->name))
			{
				return alg;
			}
		}

		return NULL;
	}
} // wtc

namespace wtc
{
namespace codegen
{

const std::string CodeGen::MainFuncName("main");
const std::string CodeGen::CtorFuncName("ctor");
const std::string CodeGen::DtorFuncName("dtor");
const std::string CodeGen::GlobalCtorFuncName("ctor$global");
const std::string CodeGen::NewObjFuncName("$new");
const std::string CodeGen::CopyObjFuncName("$copy");
const std::string CodeGen::DelObjFuncName("$del");
const std::string CodeGen::GeneratedFuncTag("$impl$");
const std::string CodeGen::AssertionFailExc("AssertionFailedException");
const std::string CodeGen::PrecondViolExc("PreconditionViolationException");
const std::string CodeGen::PostcondViolExc("PostconditionViolationException");
const std::string CodeGen::ConstraintErrExc("ConstraintErrorException");
const std::string CodeGen::PlaceholderEmptyString("");

class InstrumentProblemHandlerImpl : public InstrumentProblemHandler
{
	wtc::ProblemHandler & mProbHan;
	const std::string &   mFilename;

public:
	InstrumentProblemHandlerImpl(wtc::ProblemHandler & probhan,
		const std::string & filename)
		: InstrumentProblemHandler()
		, mProbHan(probhan)
		, mFilename(filename)
	{
		// ok
	}

	~InstrumentProblemHandlerImpl() { }

	void InstrumentNote   (const std::string & msg) { mProbHan.AddNote   (mFilename, 0, msg); }
	void InstrumentWarning(const std::string & msg) { mProbHan.AddWarning(mFilename, 0, msg); }
	void InstrumentError  (const std::string & msg) { mProbHan.AddError  (mFilename, 0, msg); }

	wtc::ProblemHandler & GetProblemHandler() { return mProbHan; }
};

CodeGen::CodeGen(wtc::Options* options, wtc::ProblemHandler* probhan, wtast::Program* program)
	: mOptions(options)
	, mProblemHandler(probhan)
	, mInstrumentProblemHandler(new InstrumentProblemHandlerImpl(*probhan, options->GetOutputFilename()))
	, mProgram(program)
	, mNextUniqueId(program->GetLastUniqueId())
	, mNextId(1)
	, mCtorPC(0)
	, mDtorPC(0)
	, mDataAddress(StartAddress)
	, mExtAddress(StartAddress)
	, mGlobalAddress(StartAddress)
	, mCodeAddress(StartAddress)
	, mProgramSize(0)
	, mGlobalSize(0)
	, mEngine(mInstrumentProblemHandler)
	, mMainBlock(0)
	, mCtorBlock(0)
	, mDtorBlock(0)
	, mExternals(0)
	, mUnifiedObjectType(0)
	, mIndexTypeU32(0)
{
	mGenFuncs["strtok"]      = &CodeGen::GenStdlib_strtok;
	mGenFuncs["empty"]       = &CodeGen::GenSeq_empty;
	mGenFuncs["seqlen"]      = &CodeGen::GenSeq_seqlen;
	mGenFuncs["seqfill"]     = &CodeGen::GenSeq_seqfill;
	mGenFuncs["subseq"]      = &CodeGen::GenSeq_subseq;
	mGenFuncs["lov"]         = &CodeGen::GenSeq_lov;
	mGenFuncs["hiv"]         = &CodeGen::GenSeq_hiv;
	mGenFuncs["lorem"]       = &CodeGen::GenSeq_lorem;
	mGenFuncs["hirem"]       = &CodeGen::GenSeq_hirem;
	mGenFuncs["loext"]       = &CodeGen::GenSeq_loext;
	mGenFuncs["hiext"]       = &CodeGen::GenSeq_hiext;
	mGenFuncs["lopop"]       = &CodeGen::GenSeq_lopop;
	mGenFuncs["hipop"]       = &CodeGen::GenSeq_hipop;
	mGenFuncs["fstat"]       = &CodeGen::GenIO_fstat;
	mGenFuncs["fopen"]       = &CodeGen::GenIO_fopen;
	mGenFuncs["feof"]        = &CodeGen::GenIO_feof;
	mGenFuncs["fread"]       = &CodeGen::GenIO_fread;
	mGenFuncs["freadchar"]   = &CodeGen::GenIO_freadchar;
	mGenFuncs["freadbool"]   = &CodeGen::GenIO_freadbool;
	mGenFuncs["freadstr"]    = &CodeGen::GenIO_freadstr;
	mGenFuncs["freadln"]     = &CodeGen::GenIO_freadln;
	mGenFuncs["freadi8"]     = &CodeGen::GenIO_freadi8;
	mGenFuncs["freadu8"]     = &CodeGen::GenIO_freadu8;
	mGenFuncs["freadi16"]    = &CodeGen::GenIO_freadi16;
	mGenFuncs["freadu16"]    = &CodeGen::GenIO_freadu16;
	mGenFuncs["freadi32"]    = &CodeGen::GenIO_freadi32;
	mGenFuncs["freadu32"]    = &CodeGen::GenIO_freadu32;
	mGenFuncs["freadf32"]    = &CodeGen::GenIO_freadf32;
	mGenFuncs["freadi64"]    = &CodeGen::GenIO_freadi64;
	mGenFuncs["freadu64"]    = &CodeGen::GenIO_freadu64;
	mGenFuncs["freadf64"]    = &CodeGen::GenIO_freadf64;
	mGenFuncs["fwrite"]      = &CodeGen::GenIO_fwrite;
	mGenFuncs["fwriteln"]    = &CodeGen::GenIO_fwriteln;
	mGenFuncs["opendir"]     = &CodeGen::GenIO_opendir;
	mGenFuncs["readdir"]     = &CodeGen::GenIO_readdir;
	mGenFuncs["dirgetname"]  = &CodeGen::GenIO_dirgetname;
	mGenFuncs["dirgettype"]  = &CodeGen::GenIO_dirgettype;
	mGenFuncs["succ"]        = &CodeGen::GenEnum_succ;
	mGenFuncs["pred"]        = &CodeGen::GenEnum_pred;
	mGenFuncs["ordinal"]     = &CodeGen::GenEnum_ordinal;
	mGenFuncs["tos"]         = &CodeGen::GenEnum_tos;
	mGenFuncs["=="]          = &CodeGen::GenEnum_eq;
	mGenFuncs["!="]          = &CodeGen::GenEnum_ne;
	mGenFuncs["<"]           = &CodeGen::GenEnum_lt;
	mGenFuncs["<="]          = &CodeGen::GenEnum_le;
	mGenFuncs[">"]           = &CodeGen::GenEnum_gt;
	mGenFuncs[">="]          = &CodeGen::GenEnum_ge;

	ProcessMetadata();

	Selector<wtast::MainBlock, false> mains(program);
	if(1!=mains.size())
	{
		AddError(program, "The number of 'main' blocks must be one.");
		return;
	}

	Selector<wtast::ConstructorBlock, false> ctors(program);
	if(1<ctors.size())
	{
		AddError(program, "At most one constructor block can be defined.");
		return;
	}

	Selector<wtast::DestructorBlock, false> dtors(program);
	if(1<dtors.size())
	{
		AddError(program, "At most one destructor block can be defined.");
		return;
	}

	GlobalBlockSelector globals(program);

	mMainBlock = *mains.begin();
	if(!ctors.empty()) { mCtorBlock = *ctors.begin(); }
	if(!dtors.empty()) { mDtorBlock = *dtors.begin(); }

	PopulateFunctions();
	CollectGlobals(program);
	DataSection* const data_section = CollectData();
	if(!data_section)
	{
		return;
	}
	mExtAddress = (mDataAddress + (data_section->GetSize() + SectionPadding - 1)) & (-SectionPadding); // PARENT?
	mGlobalAddress = mExtAddress;
	mCodeAddress = mExtAddress;

	ExternalsSection* extsec = CollectDeclarations();
	if(extsec)
	{
		mGlobalAddress = (mExtAddress + (extsec->GetSize() + SectionPadding - 1)) & (-SectionPadding); // PARENT?
		mCodeAddress = mGlobalAddress;
		RegisterSection(extsec);
	}

	{
		const std::set<std::string> & hashalgs = options->HashAlgs();

		wt_bo bo = (mOptions->BigEndian() ? WTBO_BIG : WTBO_LITTLE);

		for(std::set<std::string>::const_iterator it=hashalgs.begin(),end=hashalgs.end();it!=end;++it)
		{
			const char* const name = it->c_str();

			const struct wtr_hashalg* const alg = find_hash_alg(name);

			if(NULL!=alg)
			{
				RegisterSection(new IntegritySection(*mProblemHandler,
					GenName("integrity"), alg, bo));
			}
			else
			{
				AddError(program, std::string("No such hash algorithm: \"") + name + "\".");
			}
		}
	}

	if(mCtorBlock || !globals.empty())
	{
		mEngine.AddGlobal(CtorFuncName);

		bool was_in_global = false;

		for(GlobalBlockSelector::iterator it=globals.begin(),end=globals.end();it!=end;++it)
		{
			if(was_in_global)
			{
				mEngine.RemoveLastInstrument();
			}

			was_in_global = true;

			GenerateCode(*this, mEngine, data_section, extsec, *it, VISIT_GLOBALCTOR);
		}

		if(mEngine.LastIsWith(WT_TOMB_OP_LEAVE))
		{
			mEngine.RemoveLastInstrument();
		}

		if(mCtorBlock)
		{
			GenerateCode(*this, mEngine, data_section, extsec, mCtorBlock);
		}
		else
		{
			mEngine.AddSingle(WT_TOMB_OP_LEAVE);
		}
	}

	if(mGlobalSize)
	{
		GlobalSection* const globsec = new GlobalSection(*mProblemHandler,
			GenName("global"), mGlobalAddress, mGlobalSize);

		if(globsec)
		{
			mCodeAddress = (mGlobalAddress + (globsec->GetSize() + SectionPadding - 1)) & (-SectionPadding); // PARENT?
			RegisterSection(globsec);
		}
		else
		{
			AddError(program, "Out of memory (can not allocate GlobalSection)");
			return;
		}
	}

	mEngine.AddGlobal(MainFuncName);
	GenerateCode(*this, mEngine, data_section, extsec, mMainBlock);

	if(mDtorBlock || !globals.empty())
	{
		mEngine.AddGlobal(DtorFuncName);

		bool need_leave_instrument = !mDtorBlock;

		if(mDtorBlock)
		{
			GenerateCode(*this, mEngine, data_section, extsec, mDtorBlock);

			if(!globals.empty())
			{
				if(!mEngine.LastIsWith(WT_TOMB_OP_LEAVE))
				{
					AddError(program, "Internal error: destructor block does not ends with 'leave'");
				}

				need_leave_instrument = true;
				mEngine.RemoveLastInstrument();
			}
		}

		for(GlobalBlockSelector::iterator it=globals.begin(),end=globals.end();it!=end;++it)
		{
			GenerateCode(*this, mEngine, data_section, extsec, *it, VISIT_GLOBALDTOR);
		}

		if(need_leave_instrument && !mEngine.LastIsWith(WT_TOMB_OP_LEAVE))
		{
			mEngine.AddSingle(WT_TOMB_OP_LEAVE);
		}
	}

	for(funcdefset_t::iterator it=mFunctionDefs.begin(),end=mFunctionDefs.end();it!=end;++it)
	{
		wtast::FunctionDefinition* funcdef = *it;
		std::string name = funcdef->GetFunctionSignature()->GetCanonicalName(false, true);
		mEngine.AddGlobal(name);
		GenerateCode(*this, mEngine, data_section, extsec, funcdef);
	}

	TemplateInstantiator(&CodeGen::GenGeneratedFunc, mGeneratedFunctionNames);
	TemplateInstantiator(&CodeGen::GenCtor, mCtorNames);
	TemplateInstantiator(&CodeGen::GenCopy, mCopyNames);
	TemplateInstantiator(&CodeGen::GenDtor, mDtorNames);

	DLabelSection* dlabs = 0;

	if(mOptions->LabelsTop())
	{
		dlabs = new DLabelSection(*mProblemHandler, GenName("dlabels"));
		if(dlabs)
		{
			RegisterSection(dlabs);
		}
		else
		{
			AddError(program, "Out of memory (can not allocate DLabelSection)");
			return;
		}
	}

	std::string export_section_name = GenName("export");

	CodeSection* const code = mEngine.Compile(GenName("code"), mCodeAddress,
		dlabs, mOptions->LabelsAll());
	if(!code)
	{
		AddError(program, "Failed to generate code");
		return;
	}

	if(!mExportedFunctions.empty())
	{
		ExportSection* exp = new ExportSection(*mProblemHandler,
			export_section_name);

		if(!exp)
		{
			AddError(program, "Not enough memory to allocate export section.");
			return;
		}

		for(funcsigvec_t::iterator it=mExportedFunctions.begin(),end=mExportedFunctions.end();it!=end;++it)
		{
			wtast::FunctionSignature* sig = *it;
			wt_u32                    pc  = code->GetFunctionPC(sig->GetCanonicalName(false, true));

			if(0==pc)
			{
				std::ostringstream ss;
				ss << "Failed to resolve exported function '" << sig->GetCanonicalName(true, true) << "'";
				AddError(program, ss.str());
				return;
			}

			exp->AddFunctionExport(sig, pc);
		}

		RegisterSection(exp);
	}

	RegisterSection(code);

	mProgramSize = (mCodeAddress + (code->GetSize() + SectionPadding - 1)) & (-SectionPadding); // PARENT?

	mCtorPC = code->GetFunctionPC(CtorFuncName);
	mDtorPC = code->GetFunctionPC(DtorFuncName);

	mMainThread.th_type   = WTBTT_USER;
	mMainThread.th_flag   = 0;
	mMainThread.th_cssize = mMainBlock->GetCSS();
	mMainThread.th_dssize = mMainBlock->GetDSS();
	mMainThread.th_pcbase = code->GetFunctionPC(MainFuncName);
}

CodeGen::~CodeGen()
{
	while(mSections.size())
	{
		delete mSections.begin()->second;
		mSections.erase(mSections.begin());
	}

	while(mAllocatedAstNodes.size())
	{
		delete mAllocatedAstNodes.back();
		mAllocatedAstNodes.pop_back();
	}
}

void CodeGen::TemplateInstantiator(GenFunc_FT genfunc, map_typename_functype_t & source)
{
	map_typename_functype_t done;

	while(done.size()!=source.size())
	{
		map_typename_functype_t copy = source;

		for(map_typename_functype_t::iterator it=copy.begin(),end=copy.end();it!=end;++it)
		{
			if(done.end()==done.find(it->first))
			{
				if(!mExternals || !mExternals->HasFunc(it->second.funcname))
				{
					(this->*genfunc)(it->second);
				}

				done.insert(*it);
			}
		}
	}
}

void CodeGen::TemplateInstantiator(GenFunc_S genfunc, map_uniqueid_funcsig_t & source)
{
	for(map_uniqueid_funcsig_t::iterator it=source.begin(),end=source.end();it!=end;++it)
	{
		(this->*genfunc)(it->second);
	}
}

void CodeGen::RegisterGeneratedFunction(wtast::FunctionSignature* sig)
{
	mGeneratedFunctionNames[sig->GetUniqueId()] = sig;
}

void CodeGen::AddNote(wtast::AstNode* astnode, const std::string & msg)
{
	// TODO
	AddNote("?", 0, msg);
}

void CodeGen::AddWarning(wtast::AstNode* astnode, const std::string & msg)
{
	// TODO
	AddWarning("?", 0, msg);
}

void CodeGen::AddError(wtast::AstNode* astnode, const std::string & msg)
{
	// TODO
	AddError("?", 0, msg);
}

void CodeGen::AddNote(const std::string & file, unsigned int line, const std::string & msg)
{
	mProblemHandler->AddNote(file, line, msg);
}

void CodeGen::AddWarning(const std::string & file, unsigned int line, const std::string & msg)
{
	mProblemHandler->AddWarning(file, line, msg);
}

void CodeGen::AddError(const std::string & file, unsigned int line, const std::string & msg)
{
	mProblemHandler->AddError(file, line, msg);
}

wt_u32 CodeGen::AlignSizeUp(wt_u32 size)
{
	if(size<sizeof(wt_u32))
	{
		return sizeof(wt_u32);
	}

	return size;
}

void CodeGen::RegisterSection(Section* sec)
{
	mSections[sec->GetName()] = sec;
}

const std::string & CodeGen::GetConstructorName(wtast::TypeInfo* typeinfo)
{
	wtast::Type* type = static_cast<wtast::Type*>(typeinfo);
	const std::string & type_name = type->GetName();

	if(!type->IsDefaultConstructible())
	{
		std::ostringstream ss;

		ss << "Type '";
		ss << type_name;
		ss << "' is not default constructible.";

		AddError(typeinfo, ss.str());
		return *mUnconstructibleTypes.insert(std::string("unconstructible$") + type_name).first;
	}

	map_typename_functype_t::iterator it = mCtorNames.find(type_name);
	if(mCtorNames.end()!=it)
	{
		return it->second.funcname;
	}

	if(!type->IsObject())
	{
		AddError(typeinfo, std::string("No constructor function available for type '") + type_name + "'");
		return PlaceholderEmptyString;
	}

	/* if constructor is used, generate the destructor as well */
	GetDestructorName(type);

	mCtorNames[type_name] = FuncnameType(NewObjFuncName + "(" + type_name + "):" + type_name, type);
	return mCtorNames.find(type_name)->second.funcname;
}

const std::string & CodeGen::GetDestructorName(wtast::TypeInfo* typeinfo)
{
	wtast::Type* type = static_cast<wtast::Type*>(typeinfo);
	const std::string & type_name = type->GetName();
	map_typename_functype_t::iterator it = mDtorNames.find(type_name);
	if(mDtorNames.end()!=it)
	{
		return it->second.funcname;
	}

	if(!type->IsObject())
	{
		AddError(typeinfo, std::string("No destructor function available for type '") + type_name + "'");
		return PlaceholderEmptyString;
	}

	mDtorNames[type_name] = FuncnameType(DelObjFuncName + "(" + type_name + ")", type);
	return mDtorNames.find(type_name)->second.funcname;
}

const std::string & CodeGen::GetCopyName(wtast::TypeInfo* typeinfo)
{
	wtast::Type* type = static_cast<wtast::Type*>(typeinfo);
	const std::string & type_name = type->GetName();
	map_typename_functype_t::iterator it = mCopyNames.find(type_name);
	if(mCopyNames.end()!=it)
	{
		return it->second.funcname;
	}

	if(!type->IsObject())
	{
		AddError(typeinfo, std::string("No copy function available for type '") + type_name + "'");
		return PlaceholderEmptyString;
	}

	GetDestructorName(type); // ha masolunk, akkor torolni is kell majd

	mCopyNames[type_name] = FuncnameType(CopyObjFuncName + "(" + type_name + "):" + type_name, type);
	return mCopyNames.find(type_name)->second.funcname;
}

void CodeGen::GenCtor(FuncnameType & ft)
{
	wtast::Type* type = ft.type;

	mEngine.AddGlobal(ft.funcname);

	if(wtast::AstNode::BASICTYPE==type->GetKind())
	{
		const wt_u32 type_id = static_cast<wtast::BasicType*>(type)->GetBasicTypeId();

		if(WTT_STRING==type_id)
		{
			mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 0);
			mEngine.AddSingle(WT_TOMB_OP_STRN);
		}
		else
		if(WTT_FILE==type_id || WTT_DIR==type_id)
		{
			AddError(mProgram, "Internal error: 'file' and 'dir' types are not default constructible");
		}
	}
	else
	if(wtast::AstNode::RECORDTYPE==type->GetKind())
	{
		wtast::RecordType* rec = static_cast<wtast::RecordType*>(type);

		mEngine.AddSingle(WT_TOMB_OP_DVTA_I, rec->GetLoadSize());
		mEngine.MakeRECN(rec->GetTotalSize());
		mEngine.MakePVT(0);
		mEngine.MakeST(rec->GetLoadSize());

		for(std::size_t i=0,n=rec->GetNumberOfFields();i!=n;++i)
		{
			wtast::RecordField* field = rec->GetField(i);
			DefaultInitializeRecordField(rec, 0, field);
		}

		mEngine.MakePVT(0);
		mEngine.MakeLD(rec->GetLoadSize());
		mEngine.AddSingle(WT_TOMB_OP_DVTF);
	}
	else
	if(wtast::AstNode::DATATYPE==type->GetKind())
	{
		wtast::DataType*     dt   = static_cast<wtast::DataType*>(type);
		wtast::DataTypeCons* cons = dt->GetDefaultCons();

		if(!cons)
		{
			AddError(dt, std::string("No default constructor for datatype '") + type->GetName() + "'");
			return;
		}

		if(0!=cons->GetNumberOfTypes())
		{
			AddError(dt, std::string("Default constructor '") + type->GetName() + "::" + cons->GetName() + "' has type arguments.");
			return;
		}

		mEngine.MakeRECN(cons->GetDataSize());
		// set ConsId
		mEngine.MakePushImm(cons->GetConsId());
		mEngine.MakePDSP(sizeof(wt_u32));
		mEngine.MakeLD(dt->GetLoadSize());
		mEngine.MakeRECFR(0);
		mEngine.MakeST(sizeof(wt_u32));
		// set RefCount
		mEngine.MakePushImm(static_cast<wt_u32>(1));
		mEngine.MakePDSP(sizeof(wt_u32));
		mEngine.MakeLD(dt->GetLoadSize());
		mEngine.MakeRECFR(sizeof(wt_u32));
		mEngine.MakeST(sizeof(wt_u32));
	}
	else
	if(wtast::AstNode::SEQTYPE==type->GetKind())
	{
		GenSeqCtor(std::string(), static_cast<wtast::SeqType*>(type), static_cast<wtast::SeqType*>(type)->GetElementType());
	}
	else
	{
		AddError(mProgram, std::string("Constructor can not be generated for type '") + type->GetName() + "'");
		return;
	}

	GenGenericDtorAndLeave(type);
}

void CodeGen::GenDtor(FuncnameType & ft)
{
	wtast::Type* type = ft.type;

	mEngine.AddGlobal(ft.funcname);

	if(wtast::AstNode::BASICTYPE==type->GetKind())
	{
		const wt_u32 type_id = static_cast<wtast::BasicType*>(type)->GetBasicTypeId();

		if(WTT_STRING==type_id)
		{
			mEngine.AddSingle(WT_TOMB_OP_STRD);
		}
		else
		if(WTT_FILE==type_id)
		{
			GenIO_FileDtor();
		}
		else
		if(WTT_DIR==type_id)
		{
			GenIO_DirDtor();
		}
	}
	else
	if(wtast::AstNode::RECORDTYPE==type->GetKind())
	{
		wtast::RecordType* rec = static_cast<wtast::RecordType*>(type);

		for(std::size_t i=0,n=rec->GetNumberOfFields();i!=n;++i)
		{
			wtast::RecordField* field = rec->GetField(i);
			wtast::Type*        type  = field->GetType();

			if(type->IsObject())
			{
				mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
				mEngine.MakeRECFR(field->GetOffset());
				mEngine.MakeLD(type->GetLoadSize());

				std::string lab_skip = mEngine.GenLabel();
				std::string lab_after = mEngine.GenLabel();
				mEngine.MakeDUP(AlignSizeUp(type->GetLoadSize()));
				if(type->GetLoadSize()!=sizeof(wt_u32)) { mEngine.AddSingle(WT_TOMB_OP_BOR); }
				mEngine.AddJFI(lab_skip);
				GenDestructorCall(type);
				mEngine.AddJmpI(lab_after);
				mEngine.AddLabel(lab_skip);
				mEngine.MakePOP(type->GetLoadSize());
				mEngine.AddLabel(lab_after);
			}
		}

		mEngine.AddSingle(WT_TOMB_OP_RECD);
	}
	else
	if(wtast::AstNode::DATATYPE==type->GetKind())
	{
		wtast::DataType* dt = static_cast<wtast::DataType*>(type);

		const std::string del_label = mEngine.GenLabel();
		const std::string end_label = mEngine.GenLabel();
		const std::string keep_label = mEngine.GenLabel();

		// decrement the refcounter

		// [Addr]
		mEngine.MakeDUP(sizeof(wt_vmptr));
		// [Addr, Addr]
		mEngine.MakeRECFR(sizeof(wt_u32));
		// [Addr, &RefCount]
		mEngine.MakeLD(sizeof(wt_u32));
		// [Addr, RefCount]
		mEngine.AddSingle(WT_TOMB_OP_SUB_I, 1);
		// [Addr, NewRefCount]
		mEngine.MakePDSP(sizeof(wt_u32));
		// [Addr, NewRefCount, &Addr]
		mEngine.MakeLD(sizeof(wt_vmptr));
		// [Addr, NewRefCount, Addr]
		mEngine.MakeRECFR(sizeof(wt_u32));
		// [Addr, NewRefCount, &RefCount]
		mEngine.MakeST(sizeof(wt_u32));
		// [Addr]

		// load refcounter

		// [Addr]
		mEngine.MakeDUP(sizeof(wt_vmptr));
		// [Addr, Addr]
		mEngine.MakeRECFR(sizeof(wt_u32));
		// [Addr, &RefCount]
		mEngine.MakeLD(sizeof(wt_u32));
		// [Addr, RefCount]
		mEngine.AddJFI(del_label);
		// [Addr]
		mEngine.AddSingle(WT_TOMB_OP_POP_I, sizeof(wt_vmptr));
		// []
		mEngine.AddJmpL(keep_label);

		// really delete it

		mEngine.AddLabel(del_label);
		// [Addr]
		mEngine.MakeDUP(sizeof(wt_vmptr));
		// [Addr, Addr]
		mEngine.MakeRECFR(0);
		// [Addr, &ConsId]
		mEngine.MakeLD(sizeof(wt_u32));
		// [Addr, ConsId]

		for(std::size_t i=0,n=dt->GetNumberOfConses();i!=n;++i)
		{
			wtast::DataTypeCons* cons = dt->GetCons(i);

			bool have_obj = false;

			for(std::size_t i=0,n=cons->GetNumberOfTypes();!have_obj && i!=n;++i)
			{
				have_obj = cons->GetType(i)->IsObject();
			}

			if(!have_obj)
			{
				continue;
			}

			const std::string nextcons_label = mEngine.GenLabel();

			// [Addr, ConsId]
			mEngine.MakeDUP(sizeof(wt_u32));
			// [Addr, ConsId, ConsId]
			mEngine.MakePushImm(cons->GetConsId());
			// [Addr, ConsId, ConsId, CurrId]
			mEngine.AddSingle(WT_TOMB_OP_CMP);
			// [Addr, ConsId, Result]

			// JT is OK, because we get 0 if CMP-ing equal values
			if((cons->GetNumberOfTypes()*10+1)<128)
			{
				mEngine.AddJTI(nextcons_label);
			}
			else
			{
				mEngine.AddJTL(nextcons_label);
			}

			// [Addr, ConsId]
			for(std::size_t i=0,n=cons->GetNumberOfTypes();i!=n;++i)
			{
				wtast::Type* type = cons->GetType(i);

				if(type->IsObject())
				{
					// [Addr, ConsId]
					mEngine.MakePDSP(sizeof(wt_u32));
					// [Addr, ConsId, &Addr]
					mEngine.MakeLD(sizeof(wt_vmptr));
					// [Addr, ConsId, Addr]
					mEngine.MakeRECFR(cons->GetOffset(i));
					// [Addr, ConsId, &Object]
					mEngine.MakeLD(type->GetLoadSize());
					// [Addr, ConsId, Object]

					std::string lab_skip = mEngine.GenLabel();
					std::string lab_after = mEngine.GenLabel();
					// [Addr, ConsId, Object]
					mEngine.MakeDUP(AlignSizeUp(type->GetLoadSize()));
					// [Addr, ConsId, Object, Object]
					if(type->GetLoadSize()!=sizeof(wt_u32)) { mEngine.AddSingle(WT_TOMB_OP_BOR); }
					mEngine.AddJFI(lab_skip);
					// [Addr, ConsId, Object]
					GenDestructorCall(type);
					mEngine.AddJmpI(lab_after);
					mEngine.AddLabel(lab_skip);
					// [Addr, ConsId, Object]
					mEngine.MakePOP(type->GetLoadSize());
					// [Addr, ConsId]
					mEngine.AddLabel(lab_after);
				}
			}

			mEngine.AddJmpL(end_label);
			mEngine.AddLabel(nextcons_label);
		}

		// default case for all conses without object argument

		mEngine.AddLabel(end_label);
		// [Addr, ConsId]
		mEngine.AddSingle(WT_TOMB_OP_POP_I, sizeof(wt_u32));
		// [Addr]
		mEngine.AddSingle(WT_TOMB_OP_RECD);
		// []
		mEngine.AddLabel(keep_label);
	}
	else
	if(wtast::AstNode::SEQTYPE==type->GetKind())
	{
		GenSeqDtor(std::string(), static_cast<wtast::SeqType*>(type), static_cast<wtast::SeqType*>(type)->GetElementType());
	}
	else
	{
		AddError(mProgram, std::string("Destructor can not be generated for type '") + type->GetName() + "'");
	}

	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenCopy(FuncnameType & ft)
{
	wtast::Type* type = ft.type;

	mEngine.AddGlobal(ft.funcname);

	if(wtast::AstNode::BASICTYPE==type->GetKind())
	{
		const wt_u32 type_id = static_cast<wtast::BasicType*>(type)->GetBasicTypeId();

		if(WTT_STRING==type_id)
		{
			mEngine.AddSingle(WT_TOMB_OP_STRC);
		}
		else
		if(WTT_FILE==type_id)
		{
			GenIO_FileCopy();
		}
		else
		if(WTT_DIR==type_id)
		{
			GenIO_DirCopy();
		}
	}
	else
	if(wtast::AstNode::RECORDTYPE==type->GetKind())
	{
		wtast::RecordType* rec = static_cast<wtast::RecordType*>(type);

		const wt_u32 offs_new = 0;
		const wt_u32 offs_old = rec->GetLoadSize();

		mEngine.AddSingle(WT_TOMB_OP_DVTA_I, 2*rec->GetLoadSize());
		mEngine.MakePVT(offs_old);
		mEngine.MakeST(rec->GetLoadSize());
		mEngine.MakeRECN(rec->GetTotalSize());
		mEngine.MakePVT(offs_new);
		mEngine.MakeST(rec->GetLoadSize());

		for(std::size_t i=0,n=rec->GetNumberOfFields();i!=n;++i)
		{
			wtast::RecordField* field  = rec->GetField(i);
			wtast::Type*        type   = field->GetType();
			const bool          is_obj = type->IsObject();

			mEngine.MakePVT(offs_old);
			mEngine.MakeLD(rec->GetLoadSize());
			mEngine.MakeRECFR(field->GetOffset());
			mEngine.MakeLD(type->GetLoadSize());
			if(is_obj)
			{
				GenCopyCall(type);
			}
			mEngine.MakePVT(offs_new);
			mEngine.MakeLD(rec->GetLoadSize());
			mEngine.MakeRECFR(field->GetOffset());
			if(is_obj)
			{
				mEngine.AddSingle(WT_TOMB_OP_TDR);
			}
			mEngine.MakeST(type->GetLoadSize());
		}

		mEngine.MakePVT(offs_new);
		mEngine.MakeLD(rec->GetLoadSize());
		mEngine.AddSingle(WT_TOMB_OP_DVTF);
	}
	else
	if(wtast::AstNode::DATATYPE==type->GetKind())
	{
		// increment the refcounter

		// [Addr]
		mEngine.MakeDUP(sizeof(wt_vmptr));
		// [Addr, Addr]
		mEngine.MakeRECFR(sizeof(wt_u32));
		// [Addr, &RefCount]
		mEngine.MakeLD(sizeof(wt_u32));
		// [Addr, RefCount]
		mEngine.AddSingle(WT_TOMB_OP_ADD_I, 1);
		// [Addr, NewRefCount]
		mEngine.MakePDSP(sizeof(wt_u32));
		// [Addr, NewRefCount, &Addr]
		mEngine.MakeLD(sizeof(wt_vmptr));
		// [Addr, NewRefCount, Addr]
		mEngine.MakeRECFR(sizeof(wt_u32));
		// [Addr, NewRefCount, &RefCount]
		mEngine.MakeST(sizeof(wt_u32));
		// [Addr]
	}
	else
	if(wtast::AstNode::SEQTYPE==type->GetKind())
	{
		wtast::Type* elemtype = static_cast<wtast::SeqType*>(type)->GetElementType();

		if(elemtype->IsObject())
		{
			GenSeqCopy(std::string(), static_cast<wtast::SeqType*>(type), elemtype);
		}
		else
		{
			std::string impl_name = CreateImplName(CopyObjFuncName, elemtype->GetLoadSize());
			bool        gen_impl  = mGeneratedFuncImplNames.end()==mGeneratedFuncImplNames.find(impl_name);

			mEngine.AddJmpL(impl_name);

			if(gen_impl)
			{
				GenSeqCopy(impl_name, static_cast<wtast::SeqType*>(type), elemtype);
			}
		}
	}
	else
	{
		AddError(mProgram, std::string("Copy function can not be generated for type '") + type->GetName() + "'");
	}

	GenGenericDtorAndLeave(type);
}

void CodeGen::GenGenericDtorAndLeave(wtast::Type* type)
{
	// generic part: add dtor code

	mEngine.AddSingle(WT_TOMB_OP_DUP_I, AlignSizeUp(type->GetLoadSize()));
	mEngine.AddMkfp(GetDestructorName(type));
	mEngine.AddSingle(WT_TOMB_OP_TPU);

	// leave function

	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenGeneratedFunc(wtast::FunctionSignature* sig)
{
	const std::string & name = sig->GetName();

	map_str_genfunc_t::iterator genfunc_it = mGenFuncs.find(name);
	if(mGenFuncs.end()!=genfunc_it)
	{
		(this->*genfunc_it->second)(sig);
		return;
	}

	if(1==sig->GetArgc() && 1==sig->GetRetc() && wtast::AstNode::ENUMTYPE==sig->GetRet(0)->GetKind() &&
		name==std::string("to")+static_cast<wtast::EnumType*>(sig->GetRet(0))->GetName())
	{
		const wt_u32 type_id = sig->GetArg(0)->GetType()->GetTypeId();

		if(WTT_U32==type_id)
		{
			GenEnum_toEnum_u32(sig);
			return;
		}
		else
		if(WTT_STRING==type_id)
		{
			GenEnum_toEnum_string(sig);
			return;
		}
	}

	AddError(mProgram, std::string("Internal error: no generator for function ") + sig->GetCanonicalName(true, true));
}

void CodeGen::GenGenericCall(const std::string & fullname)
{
	if(mExternals && mExternals->HasFunc(fullname))
	{
		const wt_u32 addr = mExternals->GetFuncAddr(fullname);

		if(ExternalsSection::BadAddr!=addr)
		{
			mEngine.MakeDLE(addr);
		}
		else
		{
			AddError(mProgram, std::string("Could not resolve function: '") +
				fullname + "'");
		}
	}
	else
	{
		mEngine.AddEnter(fullname);
	}
}

void CodeGen::GenGenericMkfp(const std::string & fullname)
{
	if(mExternals && mExternals->HasFunc(fullname))
	{
		const wt_u32 addr = mExternals->GetFuncAddr(fullname);

		if(ExternalsSection::BadAddr!=addr)
		{
			mEngine.MakePBA(addr);
			mEngine.AddSingle(WT_TOMB_OP_LDFP);
		}
		else
		{
			AddError(mProgram, std::string("Could not resolve function: '") +
				fullname + "'");
		}
	}
	else
	{
		mEngine.AddMkfp(fullname);
	}
}

void CodeGen::GenConstructorCall(wtast::Type* type)
{
	const std::string & name = GetConstructorName(type);

	if(!name.empty())
	{
		GenGenericCall(name);
	}
}

void CodeGen::GenCopyCall(wtast::Type* type)
{
	const std::string & name = GetCopyName(type);

	if(!name.empty())
	{
		GenGenericCall(name);
	}
}

void CodeGen::GenDestructorCall(wtast::Type* type)
{
	const std::string & name = GetDestructorName(type);

	if(!name.empty())
	{
		GenGenericCall(name);
	}
}

void CodeGen::GenDestructorMkfp(wtast::Type* type)
{
	const std::string & name = GetDestructorName(type);

	if(!name.empty())
	{
		GenGenericMkfp(name);
	}
}

void CodeGen::DefaultInitializeRecordField(wtast::RecordType* rec, wt_u32 pvt_offs, wtast::RecordField* field)
{
	wtast::Type* type = field->GetType();

	if(!type->IsDefaultConstructible())
	{
		std::ostringstream ss;

		ss << "Record field '";
		ss << field->GetName();
		ss << "' is not default constructible.";

		AddError(field, ss.str());

		return;
	}

	if(type->IsObject())
	{
		GenConstructorCall(type);
		mEngine.AddSingle(WT_TOMB_OP_TDR);
	}
	else
	if(sizeof(wt_u32)>=type->GetLoadSize())
	{
		mEngine.AddSingle(WT_TOMB_OP_PWZ_I);
	}
	else
	if(sizeof(wt_u64)==type->GetLoadSize())
	{
		mEngine.AddSingle(WT_TOMB_OP_PWZL_I);
	}
	else
	{
		AddError(mProgram, std::string("Internal error: invalid loadsize of record field '")
			+ rec->GetName() + "." + field->GetName() + "'");
	}

	mEngine.MakePVT(pvt_offs);
	mEngine.MakeLD(rec->GetLoadSize());
	mEngine.MakeRECFR(field->GetOffset());
	mEngine.MakeST(type->GetLoadSize());
}

void CodeGen::Save(const std::string & filename)
{
	File file(*mProblemHandler, filename,
		mOptions->BigEndian() ^ mOptions->IsHostBigEndian());

	if(mOptions->Shebang())
	{
		file.WriteShebang("#! /usr/bin/env wt");
	}

	typedef std::vector<std::pair<Section*, wtb_shdr> > shdrs_t;
	shdrs_t shdrs;
	wt_u32  size = sizeof(wtb_hdr)+sizeof(wtb_thdr)+sizeof(wtb_shdr)*(1+mSections.size());

	for(sections_t::iterator it=mSections.begin(),end=mSections.end();it!=end;++it)
	{
		if(it->second->IsEmpty())
		{
			size -= sizeof(wtb_shdr);
		}
	}

	{
		wtb_shdr shdr;
		memset(&shdr, 0, sizeof(wtb_shdr));
		shdrs.push_back(std::make_pair(reinterpret_cast<Section*>(0), shdr));
	}

	for(sections_t::iterator it=mSections.begin(),end=mSections.end();it!=end;++it)
	{
		Section* sec = it->second;
		wtb_shdr shdr;

		if(sec->IsEmpty()) { continue; }

		shdr.sh_type = sec->GetSectionType();
		shdr.sh_flag = sec->GetSectionFlag();
		shdr.sh_link = 0;
		shdr.sh_offs = size;
		shdr.sh_size = sec->GetSize();
		shdr.sh_addr = sec->GetAddr();

		shdrs.push_back(std::make_pair(sec, shdr));

		if((sec->GetSectionFlag()&(WTBSF_CLEAR|WTBSF_LOAD))==WTBSF_CLEAR)
		{
			// ha csak clear, akkor nincs adat tartalom
			continue;
		}

		size += (shdr.sh_size + SectionPadding - 1) & (-SectionPadding);
	}

	for(shdrs_t::iterator it=shdrs.begin(),end=shdrs.end();it!=end;++it)
	{
		if(it->first && !it->first->GetSectionLink().empty())
		{
			const std::string & link = it->first->GetSectionLink();
			wtb_shdr          & shdr = it->second;

			shdrs_t::iterator it2, end2;

			for(it2=shdrs.begin(),end2=shdrs.end();it2!=end2;++it2)
			{
				if(it2->first && link==it2->first->GetName())
				{
					break;
				}

				++shdr.sh_link;
			}

			if(it2==end2)
			{
				AddError(mProgram, "Section can not be linked to other section.");
				return;
			}
		}
	}

	wtb_hdr hdr;

	memset(&hdr, 0, sizeof(hdr));

	hdr.b_ident[WTBI_MAGIC0]    = WTBMAG0;
	hdr.b_ident[WTBI_MAGIC1]    = WTBMAG1;
	hdr.b_ident[WTBI_MAGIC2]    = WTBMAG2;
	hdr.b_ident[WTBI_MAGIC3]    = WTBMAG3;
	hdr.b_ident[WTBI_VERSION]   = WTBV_CURRENT;
	hdr.b_ident[WTBI_BYTEORDER] = mOptions->BigEndian() ? WTBO_BIG : WTBO_LITTLE;
	hdr.b_ident[WTBI_COMPVER0]  = (wt_u8)WT_GETMAJOR(WTR_VERSION);
	hdr.b_ident[WTBI_COMPVER1]  = (wt_u8)WT_GETMINOR(WTR_VERSION);
	hdr.b_ident[WTBI_COMPVER2]  = (wt_u8)WT_GETPATCH(WTR_VERSION);
	hdr.b_ident[WTBI_COMPVER3]  = 0;
	hdr.b_size                  = mProgramSize;
	hdr.b_shnum                 = shdrs.size();
	hdr.b_thnum                 = 1;
	hdr.b_shoff                 = sizeof(wtb_hdr)+sizeof(wtb_thdr);
	hdr.b_thoff                 = sizeof(wtb_hdr);
	hdr.b_ctor                  = mCtorPC;
	hdr.b_dtor                  = mDtorPC;

	file.Write(hdr);
	file.Write(mMainThread);

	for(shdrs_t::iterator it=shdrs.begin(),end=shdrs.end();it!=end;++it)
	{
		file.Write(it->second);
	}

	for(shdrs_t::iterator it=shdrs.begin(),end=shdrs.end();it!=end;++it)
	{
		if(it->first)
		{
			it->first->Write(file);

			if((it->second.sh_flag&(WTBSF_CLEAR|WTBSF_LOAD))==WTBSF_CLEAR)
			{
				// ha csak clear, akkor nincs adat tartalom
				continue;
			}

			wt_u32 unpadded_size = it->second.sh_size;
			wt_u32 padded_size   = (unpadded_size + SectionPadding - 1) & (-SectionPadding);

			while(unpadded_size<padded_size)
			{
				file.Write(static_cast<wt_u8>(0));
				++unpadded_size;
			}
		}
	}
}

void CodeGen::ProcessMetadata()
{
	typedef Selector<wtast::Metadata, false> MetadataSelector;

	MetadataSelector mds(mProgram);
	if(!mds.empty())
	{
		MetadataSection* sec = new MetadataSection(*mProblemHandler,
			GenName("metadata"));

		if(!sec)
		{
			AddError(mProgram, "Failed to allocate memory for metadata section.");
			return;
		}

		for(MetadataSelector::iterator it=mds.begin(),end=mds.end();it!=end;++it)
		{
			sec->Add((*it)->GetKey(), (*it)->GetValue());
		}

		RegisterSection(sec);
	}
}

std::string CodeGen::GenName(const std::string & name)
{
	std::ostringstream ss;
	ss << std::setfill('0') << std::setw(NumberWidth) << (mNextId++) << '.' << name;
	return ss.str();
}

void CodeGen::PopulateFunctions()
{
	// STEP 1
	// importalt es exportalt fuggvenyek kigyujtese

	typedef Selector<wtast::DeclareBlock, false> DeclareBlockSelector;
	typedef Selector<wtast::FunctionDeclaration, false> FuncDeclSelector;
	typedef Selector<wtast::FunctionDefinition, false> FuncDefSelector;

	{
		FuncCallSelector funcallsel(mProgram);

		for(FuncCallSelector::iterator it=funcallsel.begin(),end=funcallsel.end();it!=end;++it)
		{
			wtast::FunctionSignature* const sig =
				ExtractFunctionSignature(*it);

			if(sig)
			{
				mCalledFunctions.insert(sig->GetUniqueId());
			}
		}
	}

	uniqueidset_t        impexp; // import-export UniqueIDs
	DeclareBlockSelector declblks(mProgram);

	for(DeclareBlockSelector::iterator it=declblks.begin(),end=declblks.end();it!=end;++it)
	{
		wtast::DeclareBlock* declblk = *it;
		const bool is_import = declblk->IsImport();

		if(is_import || declblk->IsExport())
		{
			FuncDeclSelector decls(declblk);

			for(FuncDeclSelector::iterator it=decls.begin(),end=decls.end();it!=end;++it)
			{
				wtast::FunctionSignature* sig = (*it)->GetFunctionSignature();
				wtparser::UniqueId unique_id = sig->GetUniqueId();
				impexp.insert(unique_id);
				mUidToFuncSig[unique_id] = sig;

				if(is_import)
				{
					mImportedFunctionIds.insert(unique_id);
				}
			}
		}
	}

	// STEP 2
	// az 'impexp' alapján a fuggveny definiciok elokeritese

	funcdefset_t    unprocessed_definitions;
	FuncDefSelector defs(mProgram);

	for(FuncDefSelector::iterator it=defs.begin(),end=defs.end();it!=end;++it)
	{
		wtast::FunctionSignature* sig       = (*it)->GetFunctionSignature();
		wtparser::UniqueId        unique_id = sig->GetUniqueId();

		if(0<impexp.count(unique_id))
		{
			unprocessed_definitions.insert(*it);
		}

		mUidToFuncDef[unique_id] = *it;

		if(IsFunctionDefined(sig->GetCanonicalName(false, false), true))
		{
			AddError(*it, "Multiple defined function.");
		}
	}

	// STEP 3
	// main, ctor es dtor hivasainak megkeresese

	{
		FuncCallSelector calls_main(mMainBlock);
		ProcessFuncCalls(unprocessed_definitions, calls_main);
		if(mCtorBlock) { FuncCallSelector calls(mCtorBlock); ProcessFuncCalls(unprocessed_definitions, calls); }
		if(mDtorBlock) { FuncCallSelector calls(mDtorBlock); ProcessFuncCalls(unprocessed_definitions, calls); }
	}

	// STEP 4
	// az 'unprocessed_definitions' elfogyasztasa, es az elert fuggvenyek eltarolasa az 'mFunctionDefs'-ben

	ProcessDefinitions(unprocessed_definitions);
}

void CodeGen::ProcessDefinitions(funcdefset_t & unprocessed_definitions)
{
	while(!unprocessed_definitions.empty())
	{
		wtast::FunctionDefinition* def = *unprocessed_definitions.begin();
		unprocessed_definitions.erase(unprocessed_definitions.begin());
		mFunctionDefs.insert(def);
		ValidateFunctionDefinition(def); // validate only the used functions

		mExportedFunctionIds.insert(def->GetFunctionSignature()->GetUniqueId());

		FuncCallSelector calls(def);
		ProcessFuncCalls(unprocessed_definitions, calls);
	}
}

void CodeGen::ValidateFunctionDefinition(wtast::FunctionDefinition* def)
{
	validator::ValidatePureProperty(mProblemHandler, def);
	validator::ValidateReturnCoverage(mProblemHandler, def);
}

void CodeGen::ProcessFuncCalls(funcdefset_t & unprocessed_definitions, FuncCallSelector & calls)
{
	for(FuncCallSelector::iterator it=calls.begin(),end=calls.end();it!=end;++it)
	{
		wtast::FunctionSignature* const sig = ExtractFunctionSignature(*it);

		if(sig)
		{
			map_uniqueid_funcdef_t::iterator it2 =
				mUidToFuncDef.find(sig->GetUniqueId());

			if(mUidToFuncDef.end()!=it2 &&
				0==mFunctionDefs.count(it2->second))
			{
				mFunctionDefs.insert(it2->second);
				unprocessed_definitions.insert(it2->second);
			}
		}
	}
}

ExternalsSection* CodeGen::CollectDeclarations()
{
	typedef Selector<wtast::DeclareBlock, false> DeclareBlockSelector;
	typedef Selector<wtast::AstNode, true, wtast::AssertStatement, wtast::PreCondition, wtast::PostCondition, wtast::RaiseStatement, wtast::EHandler, wtast::FunctionCall> SpecialStmtSelector;

	DeclareBlockSelector declblks(mProgram);
	SpecialStmtSelector specialstmts(mProgram);

	if(declblks.empty() && specialstmts.empty())
	{
		return 0;
	}

	ExternalsSection* ext = new ExternalsSection(*mProblemHandler,
		GenName("externals"), mExtAddress);

	if(!ext)
	{
		AddError(mProgram, "Out of memory in CodeGen::CollectDeclarations()");
		return 0;
	}

	std::set<std::string> referenced_exceptions;

	for(SpecialStmtSelector::iterator it=specialstmts.begin(),end=specialstmts.end();it!=end;++it)
	{
		switch((*it)->GetKind())
		{
		case wtast::AstNode::ASSERTSTATEMENT:
			if(mOptions->EnableAsserts() && !ext->HasException(AssertionFailExc))
			{
				referenced_exceptions.insert(AssertionFailExc);
			}
			break;

		case wtast::AstNode::PRECONDITION:
			if(mOptions->EnablePrecond() && !ext->HasException(PrecondViolExc))
			{
				referenced_exceptions.insert(PrecondViolExc);
			}
			break;

		case wtast::AstNode::POSTCONDITION:
			if(mOptions->EnablePostcond() && !ext->HasException(PostcondViolExc))
			{
				referenced_exceptions.insert(PostcondViolExc);
			}
			break;

		case wtast::AstNode::RAISESTATEMENT:
			referenced_exceptions.insert(static_cast<wtast::RaiseStatement*>(*it)->GetException());
			break;

		case wtast::AstNode::EHANDLER:
			referenced_exceptions.insert(static_cast<wtast::EHandler*>(*it)->GetName());
			break;

		case wtast::AstNode::FUNCTIONCALL:
			{
				wtast::FunctionSignature* sig = static_cast<wtast::FunctionCall*>(*it)->GetFunctionSignature();

				if(sig->IsGenerated() && 1==sig->GetArgc() && 1==sig->GetRetc())
				{
					if("succ"==sig->GetName() || "pred"==sig->GetName())
					{
						referenced_exceptions.insert(ConstraintErrExc);
					}
					else
					if(wtast::AstNode::ENUMTYPE==sig->GetRet(0)->GetKind())
					{
						wtast::EnumType* enum_type = static_cast<wtast::EnumType*>(sig->GetRet(0));
						std::string name = std::string("to") + enum_type->GetName();

						if(name==sig->GetName())
						{
							referenced_exceptions.insert(ConstraintErrExc);
						}
					}
				}
			}
			break;

		default: break;
		}
	}

	referenced_exceptions.erase("...");

	ImportIncludeWriter* impincwriter = 0;
	if(mOptions->GenerateImportInclude())
	{
		impincwriter = new ImportIncludeWriter(mProblemHandler, mProgram,
			mOptions->GetSourceFilename(), mOptions->GetImportFilename());

		if(!impincwriter)
		{
			AddError(mProgram,
				"Out of memory while creating ImportIncludeWriter");
			delete ext;
			return 0;
		}
	}

	std::vector<wtast::FunctionSignature*> special_function_imports;
	std::vector<wtast::ExceptionDeclaration*> exception_exports;
	std::vector<wtast::ExceptionDeclaration*> exception_imports;
	std::vector<std::pair<bool, wtast::Declaration*> > externals;

	for(DeclareBlockSelector::iterator it=declblks.begin(),end=declblks.end();it!=end;++it)
	{
		wtast::DeclareBlock* declblk = *it;
		const bool is_import = declblk->IsImport();
		const bool is_export = declblk->IsExport();

		if(is_import && is_export)
		{
			AddError(declblk, "Declaration section cannot import and export at the same time.");
			continue;
		}

		for(std::size_t i=0,n=declblk->GetNumDeclarations();i!=n;++i)
		{
			wtast::Declaration* decl = declblk->GetDeclaration(i);

			bool record_import_too = false;
			bool recorded = false;

			switch(decl->GetKind())
			{
			case wtast::AstNode::EXCEPTIONDECLARATION:
				{
					wtast::ExceptionDeclaration* exc_decl = static_cast<wtast::ExceptionDeclaration*>(decl);
					const std::string &          exc_name = exc_decl->GetName();

					// lokalis kivetelek nem leteznek, azokat exportalni kell, majd importalni
					// oka: a global exception ID-ra szukseg van

					if(is_export || referenced_exceptions.count(exc_name))
					{
						referenced_exceptions.erase(exc_name);

						if(!is_import)
						{
							exception_exports.push_back(exc_decl);
						}

						if(is_export && impincwriter)
						{
							impincwriter->WriteException(exc_name);
						}

						exception_imports.push_back(exc_decl);
						ext->AddExc(exc_name);
					}
				}
				break;

			case wtast::AstNode::FUNCTIONDECLARATION:
				{
					wtast::FunctionSignature* sig = static_cast<wtast::FunctionDeclaration*>(decl)->GetFunctionSignature();

					if(is_import)
					{
						if(mCalledFunctions.count(sig->GetUniqueId()))
						{
							ext->AddFunc(sig);
							recorded = true;
						}

						if(IsFunctionDefined(sig->GetCanonicalName(false, false), true))
						{
							AddError(decl, "Import of a declared function.");
						}
					}
					if(is_export)
					{
						mExportedFunctions.push_back(sig);

						if(impincwriter)
						{
							impincwriter->WriteFunction(sig);
						}
					}
				}
				break;

			case wtast::AstNode::RECORDDECLARATION:
				{
					wtast::RecordType*  rec      = static_cast<wtast::RecordDeclaration*>(decl)->GetRecord();
					const std::string & rec_name = rec->GetName();

					if(is_import && is_export)
					{
						AddError(rec, "Record type can not be imported and exported at the same time.");
						break;
					}

					if(is_import)
					{
						if(0!=mExternalTypeSet.count(rec_name))
						{
							AddError(rec, std::string("Record '") + rec_name + "' is external, thus can not be imported.");
							break;
						}

						mExternalTypeSet.insert(rec_name);
						mImportedTypeSet.insert(rec_name);

						ext->AddType(rec);
						recorded = true;

						if(rec->IsPrivate() || rec->IsLimited())
						{
							wtast::FunctionSignature* sig;

							if(!rec->IsLimited())
							{
								sig = MakeCtorFuncSig(rec);
								if(sig)
								{
									ext->AddFunc(sig);
									special_function_imports.push_back(sig);
								}

								sig = MakeCopyFuncSig(rec);
								if(sig)
								{
									ext->AddFunc(sig);
									special_function_imports.push_back(sig);
								}
							}

							sig = MakeDtorFuncSig(rec);
							if(sig)
							{
								ext->AddFunc(sig);
								special_function_imports.push_back(sig);
							}
						}
					}
					else
					if(is_export)
					{
						if(0!=mExternalTypeSet.count(rec_name))
						{
							AddError(rec, std::string("Record '") + rec_name + "' is external, thus can not be exported.");
							break;
						}

						mExternalTypeSet.insert(rec_name);
						mExportedTypeSet.insert(rec_name);

						ext->AddType(rec);
						recorded = true;
						record_import_too = true;

						if(rec->IsPrivate() || rec->IsLimited())
						{
							GetConstructorName(rec);
							GetCopyName(rec);
							GetDestructorName(rec);

							wtast::FunctionSignature* sig;

							if(!rec->IsLimited())
							{
								sig = MakeCtorFuncSig(rec);
								if(sig) { mExportedFunctions.push_back(sig); }
								sig = MakeCopyFuncSig(rec);
								if(sig) { mExportedFunctions.push_back(sig); }
							}

							sig = MakeDtorFuncSig(rec);
							if(sig) { mExportedFunctions.push_back(sig); }
						}

						if(impincwriter)
						{
							impincwriter->WriteRecord(rec);
						}
					}
				}
				break;

			case wtast::AstNode::DATATYPEDECLARATION:
				{
					wtast::DataType*    dt      = static_cast<wtast::DataTypeDeclaration*>(decl)->GetDataType();
					const std::string & dt_name = dt->GetName();

					if(is_import)
					{
						if(0!=mExternalTypeSet.count(dt_name))
						{
							AddError(dt, std::string("Datatype '") + dt_name +
								"' is external, thus can not be imported.");
							break;
						}

						mExternalTypeSet.insert(dt_name);
						mImportedTypeSet.insert(dt_name);

						ext->AddType(dt);
						recorded = true;
					}
					else
					if(is_export)
					{
						if(0!=mExternalTypeSet.count(dt_name))
						{
							AddError(dt, std::string("Datatype '") + dt_name +
								"' is external, thus can not be exported.");
							break;
						}

						mExternalTypeSet.insert(dt_name);
						mExportedTypeSet.insert(dt_name);

						ext->AddType(dt);
						recorded = true;
						record_import_too = true;

						if(impincwriter)
						{
							impincwriter->WriteDataType(dt);
						}
					}
				}
				break;

			case wtast::AstNode::ENUMTYPEDECLARATION:
				{
					wtast::EnumType*    et      = static_cast<wtast::EnumTypeDeclaration*>(decl)->GetEnumType();
					const std::string & et_name = et->GetName();

					if(is_import)
					{
						if(0!=mExternalTypeSet.count(et_name))
						{
							AddError(et, std::string("Enum '") + et_name +
								"' is external, thus can not be imported.");
							break;
						}

						mExternalTypeSet.insert(et_name);
						mImportedTypeSet.insert(et_name);

						ext->AddType(et);
						recorded = true;
					}
					else
					if(is_export)
					{
						if(0!=mExternalTypeSet.count(et_name))
						{
							AddError(et, std::string("Enum '") + et_name +
								"' is external, thus can not be exported.");
							break;
						}

						mExternalTypeSet.insert(et_name);
						mExportedTypeSet.insert(et_name);

						ext->AddType(et);
						recorded = true;
						record_import_too = true;

						if(impincwriter)
						{
							impincwriter->WriteEnumType(et);
						}
					}
				}
				break;

			case wtast::AstNode::OPERATORDECLARATION:
				{
					wtast::OperatorDeclaration* op = static_cast<wtast::OperatorDeclaration*>(decl);

					if(mKnownOperators.end()!=mKnownOperators.find(op->GetOpName()))
					{
						AddError(op, "Multiple operator declaration found.");
						break;
					}
					else
					{
						mKnownOperators.insert(op->GetOpName());
					}

					if(is_import && is_export)
					{
						AddError(op, "Operator can not be imported and exported at the same time.");
						break;
					}

					if(is_export)
					{
						ext->AddOperator(op);
						recorded = true;

						if(impincwriter)
						{
							impincwriter->WriteOperator(op);
						}
					}
					else
					if(is_import)
					{
						ext->AddOperator(op);
						recorded = true;
					}
				}
				break;

			case wtast::AstNode::FUNCTIONTYPEDECLARATION:
				{
					wtast::FunctionType* ft      = static_cast<wtast::FunctionTypeDeclaration*>(decl)->GetFunctionType();
					const std::string &  ft_name = ft->GetName();

					if(is_import)
					{
						if(0!=mExternalTypeSet.count(ft_name))
						{
							AddError(ft, std::string("Function type '") +
								ft_name +
								"' is external, thus can not be imported.");
							break;
						}

						mExternalTypeSet.insert(ft_name);
						mImportedTypeSet.insert(ft_name);

						ext->AddType(ft);
						recorded = true;
					}
					else
					if(is_export)
					{
						if(0!=mExternalTypeSet.count(ft_name))
						{
							AddError(ft, std::string("Function type '") +
								ft_name +
								"' is external, thus can not be exported.");
							break;
						}

						mExternalTypeSet.insert(ft_name);
						mExportedTypeSet.insert(ft_name);

						ext->AddType(ft);
						recorded = true;
						record_import_too = true;

						if(impincwriter)
						{
							impincwriter->WriteFunctionType(ft);
						}
					}
				}
				break;

			default:
				AddError(decl, "Unhandled declaration kind.");
				break;
			}

			if(recorded && (is_import || is_export) && wtast::AstNode::EXCEPTIONDECLARATION!=decl->GetKind())
			{
				if(!(is_export && wtast::AstNode::FUNCTIONDECLARATION==decl->GetKind()))
				{
					// - exceptions are handled differently
					// - functions are exported elsewhere
					externals.push_back(std::make_pair(is_import, decl));

					if(record_import_too)
					{
						externals.push_back(std::make_pair(true, decl));
					}
				}
			}
		}
	}

	delete impincwriter;

	ext->Compile();
	mExternals = ext;

	ImportSection* imp_sec = 0;
	ExportSection* exp_sec = 0;

	for(std::vector<wtast::ExceptionDeclaration*>::iterator it=exception_exports.begin(),end=exception_exports.end();it!=end;++it)
	{
		if(!exp_sec)
		{
			exp_sec = new ExportSection(*mProblemHandler, GenName("export"));
		}

		wtast::ExceptionDeclaration* const exc_decl = *it;

		exp_sec->AddExceptionExport(exc_decl->GetName());
	}
	exception_exports.clear();

	for(std::vector<std::pair<bool, wtast::Declaration*> >::iterator it=externals.begin(),end=externals.end();it!=end;++it)
	{
		const bool is_import = it->first;
		wtast::Declaration* const decl = it->second;

		if(is_import)
		{
			if(!imp_sec)
			{
				if(exp_sec)
				{
					RegisterSection(exp_sec);
					exp_sec = 0;
				}

				imp_sec = new ImportSection(*mProblemHandler,
					GenName("import"));
			}

			for(std::vector<wtast::ExceptionDeclaration*>::iterator it=exception_imports.begin(),end=exception_imports.end();it!=end;++it)
			{
				wtast::ExceptionDeclaration* const exc_decl = *it;
				imp_sec->AddExceptionImport(exc_decl->GetName(), mExternals->GetExcAddr(exc_decl->GetName()));
			}
			exception_imports.clear();

			switch(decl->GetKind())
			{
			case wtast::AstNode::FUNCTIONDECLARATION:
				{
					wtast::FunctionSignature* sig = static_cast<wtast::FunctionDeclaration*>(decl)->GetFunctionSignature();

					imp_sec->AddFunctionImport(sig, mExternals->GetFuncAddr(sig->GetCanonicalName(false, true)));
				}
				break;

			case wtast::AstNode::RECORDDECLARATION:
				{
					wtast::RecordType* rec = static_cast<wtast::RecordDeclaration*>(decl)->GetRecord();

					imp_sec->AddTypeImport(rec, mExternals->GetTypeOffs(rec->GetName()));
				}
				break;

			case wtast::AstNode::DATATYPEDECLARATION:
				{
					wtast::DataType* dt = static_cast<wtast::DataTypeDeclaration*>(decl)->GetDataType();

					imp_sec->AddTypeImport(dt, mExternals->GetTypeOffs(dt->GetName()));
				}
				break;

			case wtast::AstNode::ENUMTYPEDECLARATION:
				{
					wtast::EnumType* et = static_cast<wtast::EnumTypeDeclaration*>(decl)->GetEnumType();

					imp_sec->AddTypeImport(et, mExternals->GetTypeOffs(et->GetName()));
				}
				break;

			case wtast::AstNode::OPERATORDECLARATION:
				{
					wtast::OperatorDeclaration* op = static_cast<wtast::OperatorDeclaration*>(decl);

					imp_sec->AddOperatorImport(op);
				}
				break;

			case wtast::AstNode::FUNCTIONTYPEDECLARATION:
				{
					wtast::FunctionType* ft = static_cast<wtast::FunctionTypeDeclaration*>(decl)->GetFunctionType();

					imp_sec->AddTypeImport(ft, mExternals->GetTypeOffs(ft->GetName()));
				}
				break;

			default:
				AddError(decl, "Unhandled declaration kind.");
				break;
			}
		}
		else
		{
			if(imp_sec)
			{
				for(std::vector<wtast::FunctionSignature*>::iterator it=special_function_imports.begin(),end=special_function_imports.end();it!=end;++it)
				{
					wtast::FunctionSignature* sig = *it;
					imp_sec->AddFunctionImport(sig, mExternals->GetFuncAddr(sig->GetCanonicalName(false, true)));
				}
				special_function_imports.clear();
			}

			if(!exp_sec)
			{
				if(imp_sec)
				{
					RegisterSection(imp_sec);
					imp_sec = 0;
				}

				exp_sec = new ExportSection(*mProblemHandler,
					GenName("export"));
			}

			switch(decl->GetKind())
			{
			case wtast::AstNode::RECORDDECLARATION:
				{
					wtast::RecordType* rec = static_cast<wtast::RecordDeclaration*>(decl)->GetRecord();

					exp_sec->AddTypeExport(rec, mExternals->GetTypeOffs(rec->GetName()));
				}
				break;

			case wtast::AstNode::DATATYPEDECLARATION:
				{
					wtast::DataType* dt = static_cast<wtast::DataTypeDeclaration*>(decl)->GetDataType();

					exp_sec->AddTypeExport(dt, mExternals->GetTypeOffs(dt->GetName()));
				}
				break;

			case wtast::AstNode::ENUMTYPEDECLARATION:
				{
					wtast::EnumType* et = static_cast<wtast::EnumTypeDeclaration*>(decl)->GetEnumType();

					exp_sec->AddTypeExport(et, mExternals->GetTypeOffs(et->GetName()));
				}
				break;

			case wtast::AstNode::OPERATORDECLARATION:
				{
					wtast::OperatorDeclaration* op = static_cast<wtast::OperatorDeclaration*>(decl);

					exp_sec->AddOperatorExport(op);
				}
				break;

			case wtast::AstNode::FUNCTIONTYPEDECLARATION:
				{
					wtast::FunctionType* ft = static_cast<wtast::FunctionTypeDeclaration*>(decl)->GetFunctionType();

					exp_sec->AddTypeExport(ft, mExternals->GetTypeOffs(ft->GetName()));
				}
				break;

			default:
				AddError(decl, "Unhandled declaration kind.");
				break;
			}
		}
	}

	if(!exception_imports.empty() || !special_function_imports.empty())
	{
		if(exp_sec)
		{
			RegisterSection(exp_sec);
			exp_sec = 0;
		}

		if(!imp_sec)
		{
			imp_sec = new ImportSection(*mProblemHandler, GenName("import"));
		}

		for(std::vector<wtast::ExceptionDeclaration*>::iterator it=exception_imports.begin(),end=exception_imports.end();it!=end;++it)
		{
			wtast::ExceptionDeclaration* const exc_decl = *it;
			imp_sec->AddExceptionImport(exc_decl->GetName(), mExternals->GetExcAddr(exc_decl->GetName()));
		}
		exception_imports.clear();

		for(std::vector<wtast::FunctionSignature*>::iterator it=special_function_imports.begin(),end=special_function_imports.end();it!=end;++it)
		{
			wtast::FunctionSignature* sig = *it;
			imp_sec->AddFunctionImport(sig, mExternals->GetFuncAddr(sig->GetCanonicalName(false, true)));
		}
		special_function_imports.clear();
	}

	if(exp_sec)
	{
		RegisterSection(exp_sec);
		exp_sec = 0;
	}

	if(imp_sec)
	{
		RegisterSection(imp_sec);
		imp_sec = 0;
	}

	return ext;
}

DataSection* CodeGen::CollectData()
{
	std::string data_name = GenName("data");
	std::string bos_name = GenName("bos");

	DataSection* const ds = new DataSection(*mProblemHandler, data_name,
		mDataAddress, bos_name);

	if(!ds)
	{
		AddError(mProgram, "Out of memory in CodeGen::CollectData()");
		return 0;
	}

	CollectData(ds, mMainBlock);
	if(mCtorBlock) { CollectData(ds, mCtorBlock); }
	if(mDtorBlock) { CollectData(ds, mDtorBlock); }

	GlobalBlockSelector globals(mProgram);
	for(GlobalBlockSelector::iterator it=globals.begin(),end=globals.end();it!=end;++it)
	{
		CollectData(ds, *it);
	}

	for(funcdefset_t::iterator it=mFunctionDefs.begin(),end=mFunctionDefs.end();it!=end;++it)
	{
		CollectData(ds, *it);
	}

	BosSection* const bos = ds->Compile();

	if(!bos)
	{
		AddError(mProgram, "Problem occured in CodeGen::CollectData() while compiling data");
		delete ds;
		return 0;
	}

	RegisterSection(ds);
	RegisterSection(bos);

	return ds;
}

class CollectDataVisitor : public wtast::AstVisitor
{
	DataSection* mDataSection;
	bool         mInline32;
	bool         mInline64;

public:
	CollectDataVisitor(DataSection* ds, bool in32, bool in64)
		: wtast::AstVisitor()
		, mDataSection(ds)
		, mInline32(in32)
		, mInline64(in64)
	{
		// ok
	}

	bool Visit(wtast::StringLiteral* lit)
	{
		mDataSection->Add(lit->GetLiteral());
		return false;
	}

	bool Visit(wtast::NumberLiteral* lit)
	{
		if(lit->Is64Bit())
		{
			if(!mInline64)
			{
				mDataSection->Add(lit->GetLiteralU64());
			}
		}
		else
		{
			if(!mInline32)
			{
				mDataSection->Add(lit->GetLiteralU32());
			}
		}

		return false;
	}
};

void CodeGen::CollectData(DataSection* ds, wtast::AstNode* root)
{
	CollectDataVisitor cdv(ds, mOptions->Inline32(), mOptions->Inline64());
	root->Accept(&cdv);
}

bool CodeGen::IsSafeAstNode(wtast::AstNode* node, const std::string & alloc_mode_string)
{
	if(node)
	{
		return true;
	}

	AddError(mProgram, std::string("Failed to allocate memory for ") + alloc_mode_string + ".");
	return false;
}

bool CodeGen::IsSafeRegisterAstNode(wtast::AstNode* node, const std::string & alloc_mode_string)
{
	if(IsSafeAstNode(node, alloc_mode_string))
	{
		mAllocatedAstNodes.push_back(node);
		return true;
	}

	return false;
}

wtast::Type* CodeGen::GetUnifiedObjectType()
{
	if(mUnifiedObjectType)
	{
		return mUnifiedObjectType;
	}

	mUnifiedObjectType = new wtast::BasicType("*",
		(sizeof(wt_u32)==sizeof(wt_vmptr)?WTT_U32:WTT_U64), sizeof(wt_vmptr));

	if(!IsSafeRegisterAstNode(mUnifiedObjectType, "unified object type"))
	{
		abort();
	}

	return mUnifiedObjectType;
}

wtast::Type* CodeGen::GetIndexTypeU32()
{
	if(mIndexTypeU32)
	{
		return mIndexTypeU32;
	}

	mIndexTypeU32 = new wtast::BasicType("u32", WTT_U32, sizeof(wt_u32));

	if(!IsSafeRegisterAstNode(mIndexTypeU32, "index type"))
	{
		abort();
	}

	return mIndexTypeU32;
}

wtast::FunctionSignature* CodeGen::MakeFunctionSignature(const std::string & name, bool pure, wtast::Type* arg, wtast::Type* ret)
{
	wtast::FunctionSignature* sig = new wtast::FunctionSignature(name, pure);

	if(!IsSafeRegisterAstNode(sig, "function signature"))
	{
		return 0;
	}

	if(arg)
	{
		wtast::Type* arg_copy = arg->Copy();

		if(!IsSafeAstNode(arg_copy, "function argument type"))
		{
			delete sig;
			return 0;
		}

		wtast::FunctionArgument* fa = new wtast::FunctionArgument(arg_copy, "");

		if(!IsSafeAstNode(fa, "function argument"))
		{
			delete sig;
			return 0;
		}

		sig->AddArgBack(fa);
	}

	if(ret)
	{
		wtast::Type* ret_copy = ret->Copy();

		if(IsSafeAstNode(ret_copy, "function return type"))
		{
			sig->AddRetBack(ret_copy);
		}
	}

	return sig;
}

wtast::FunctionSignature* CodeGen::MakeCtorFuncSig(wtast::Type* type)
{
	return MakeFunctionSignature(NewObjFuncName, true, type, type);
}

wtast::FunctionSignature* CodeGen::MakeDtorFuncSig(wtast::Type* type)
{
	return MakeFunctionSignature(DelObjFuncName, true, type, 0);
}

wtast::FunctionSignature* CodeGen::MakeCopyFuncSig(wtast::Type* type)
{
	return MakeFunctionSignature(CopyObjFuncName, true, type, type);
}

std::string CodeGen::CreateImplName(const std::string & funcname, wt_u32 loadsize)
{
	std::ostringstream ss;
	ss << funcname << GeneratedFuncTag << loadsize;
	return ss.str();
}

bool CodeGen::IsFunctionDefined(const std::string & fullname, bool add_as_definition)
{
	if(mDefinedFunctions.count(fullname))
	{
		return true;
	}

	if(add_as_definition)
	{
		mDefinedFunctions.insert(fullname);
	}

	return false;
}

void CodeGen::AddGlobalVariable(wtast::VariableDeclFragment* frag, wt_u32 curr_dvt_size)
{
	if(mGlobalSize<curr_dvt_size)
	{
		mGlobalSize = curr_dvt_size;
	}

	mGlobalVariables.push_back(frag);
}

void CodeGen::FillGlobalVariables(FrameStack & framestack)
{
	for(vardeclfrags_t::iterator it=mGlobalVariables.begin(),end=mGlobalVariables.end();it!=end;++it)
	{
		framestack.AddVariable(*it, true);
	}
}

wtparser::UniqueId CodeGen::GetNextUniqueId()
{
	return mNextUniqueId++;
}

class CollectVarAccessVisitor : public wtast::AstVisitor
{
	CodeGen & mCodeGen;
	bool      mInGlobalBlock;

public:
	typedef std::pair<wtast::AstNode*, wtast::FunctionSignature*>
		node_funcsig_pair_t;

	std::deque<node_funcsig_pair_t> unproc_refs;

public:
	CollectVarAccessVisitor(CodeGen & cg)
		: wtast::AstVisitor()
		, mCodeGen(cg)
		, mInGlobalBlock(false)
	{
		// ok
	}

	bool Visit(wtast::GlobalBlock*)
	{
		mInGlobalBlock = true;
		return true;
	}

	void EndVisit(wtast::GlobalBlock*)
	{
		mInGlobalBlock = false;
	}

	bool Visit(wtast::VariableDeclaration* vardecl)
	{
		if(mInGlobalBlock)
		{
			for(std::size_t i=0,n=vardecl->GetNumberOfFragments();i!=n;++i)
			{
				mCodeGen.mGlobFragToGlobDecl[vardecl->GetFragment(i)->GetUniqueId()] = vardecl;
			}
		}

		return true;
	}

	bool Visit(wtast::VariableAccess* varacc)
	{
		CodeGen::map_uniqueid_vardecl_t::iterator it = mCodeGen.mGlobFragToGlobDecl.find(varacc->GetUniqueId());

		if(mCodeGen.mGlobFragToGlobDecl.end()!=it)
		{
			mCodeGen.mReferredGlobDecls.insert(it->second);
		}

		return false;
	}

	bool Visit(wtast::FunctionDefinition* funcdef)
	{
		return
			mCodeGen.mCalledFunctions.count(funcdef->GetFunctionSignature()->GetUniqueId())
			||
			mCodeGen.mExportedFunctionIds.count(funcdef->GetFunctionSignature()->GetUniqueId())
			;
	}

	bool Visit(wtast::FunctionCall* call)
	{
		wtast::FunctionSignature* sig = call->GetFunctionSignature();

		if(0==mCodeGen.mExportedFunctionIds.count(sig->GetUniqueId())
			&& !sig->IsBuiltin() && !sig->IsGenerated()
			&& 0==mCodeGen.mImportedFunctionIds.count(sig->GetUniqueId()))
		{
			unproc_refs.push_back(node_funcsig_pair_t(call, sig));
		}

		return true;
	}

	bool Visit(wtast::MakeFunctionPointer* mkfp)
	{
		wtast::FunctionSignature* sig = mkfp->GetFunctionSignature();

		if(0==mCodeGen.mExportedFunctionIds.count(sig->GetUniqueId())
			&& !sig->IsBuiltin() && !sig->IsGenerated()
			&& 0==mCodeGen.mImportedFunctionIds.count(sig->GetUniqueId()))
		{
			unproc_refs.push_back(node_funcsig_pair_t(mkfp, sig));
		}

		return false;
	}
};

void CodeGen::CollectGlobals(wtast::AstNode* root)
{
	CollectVarAccessVisitor cvav(*this);

	for(;;)
	{
		root->Accept(&cvav);

		if(cvav.unproc_refs.empty())
		{
			break;
		}

		bool has_error = false;
		funcdefset_t unproc;

		while(!cvav.unproc_refs.empty())
		{
			CollectVarAccessVisitor::node_funcsig_pair_t & nfp =
				cvav.unproc_refs.front();

			map_uniqueid_funcdef_t::iterator it =
				mUidToFuncDef.find(nfp.second->GetUniqueId());

			if(mUidToFuncDef.end()==it)
			{
				AddError(nfp.first, std::string("Definition of '") +
					nfp.second->GetCanonicalName(false, true) +
					"' not found.");
				return;
			}

			unproc.insert(it->second);
			cvav.unproc_refs.pop_front();
		}

		ProcessDefinitions(unproc);

		if(has_error)
		{
			return;
		}
	}
}

bool CodeGen::IsReferredGlobal(wtast::VariableDeclaration* vardecl)
{
	return mReferredGlobDecls.find(vardecl) != mReferredGlobDecls.end();
}

wtast::FunctionSignature* CodeGen::ExtractFunctionSignature(wtast::Expression* expr)
{
	switch(expr->GetKind())
	{
	case wtast::AstNode::FUNCTIONCALL:
		return static_cast<wtast::FunctionCall*>(expr)->GetFunctionSignature();

	case wtast::AstNode::MAKEFUNCTIONPOINTER:
		return static_cast<wtast::MakeFunctionPointer*>(expr)->GetFunctionSignature();

	default:
		AddError(expr, "Internal error: Cannot get function signature.");
		return 0;
	}
}

} // codegen
} // wtc
